@@ -73,9 +73,40 @@ int use_post=1;
 int local_only=0;
 int poll_time_min=TCPCGI_CLIENT_POLLTIME_MIN;
 int poll_time_max=TCPCGI_CLIENT_POLLTIME_MAX;
+char *userpass=NULL;
 struct sockinfo socks_info[128]; /* Needs to be num of elements as socks_poll in main() */
 
 
+char *mimecode(void *databuf, int datalen) {
+	unsigned char buffer[3];
+	unsigned char *data=databuf;
+	char mimeabet[64]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+	char *ret;
+	int outlen, retlen;
+	int i,x=0;
+
+	outlen=(int) ((((double) datalen)*1.3333333)+0.9);
+	retlen=((int) ((((double) datalen)/3.00)+0.9))*4;
+	ret=malloc(retlen+1);
+
+	for (i=0; i<datalen; i+=3) {
+		buffer[0]=data[i];
+		buffer[1]=data[i+1];
+		buffer[2]=data[i+2];
+		ret[x++]=mimeabet[((buffer[0]&0xfc)>>2)%sizeof(mimeabet)];
+		if (x>=outlen) break;
+		ret[x++]=mimeabet[(((buffer[0]&0x3)<<4)|((buffer[1]&0xf0)>>4))%sizeof(mimeabet)];
+		if (x>=outlen) break;
+		ret[x++]=mimeabet[(((buffer[1]&0xf)<<2)|((buffer[2]&0xc0)>>6))%sizeof(mimeabet)];
+		if (x>=outlen) break;
+		ret[x++]=mimeabet[(buffer[2]&0x3f)%sizeof(mimeabet)];
+		if (x>=outlen) break;
+	}
+	while (x!=retlen) ret[x++]='=';
+	ret[x]='\0';
+	return(ret);
+}
+
 char *hexcode(void *data, int datalen) {
 	char hexabet[]="0123456789abcdef";
 	unsigned char *data_num=data;
@@ -103,7 +134,7 @@ char *hexcode(void *data, int datalen) {
 		/* This is no longer used in favor of a more leanient algorithm. */
 		if (!isalnum(data_num[i]) && data_num[i]!='/' && data_num[i]!='.' && data_num[i]!='~' && data_num[i]!='+') {
 #endif
-			ret[destpos++]='%';
+			ret[destpos++]='#';
 			ret[destpos++]=hexabet[(data_num[i]>>4)&0xf];
 			ret[destpos++]=hexabet[data_num[i]&0xf];
 			continue;
@@ -119,12 +150,16 @@ char *hexcode(void *data, int datalen) {
 	return(ret);
 }
 
-void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd, char *session, char *dest, void *data, int *datalen, signed int *status, int proto_type) {
+void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *http_userpass, char *cmd, char *session, char *dest, void *data, int *datalen, signed int *status, int proto_type) {
 	FILE *fp;
-	char *hex_cmd=NULL, *hex_session=NULL, *hex_data=NULL, *hex_path=NULL, *hex_dest=NULL;
+	char *hex_cmd=NULL, *hex_session=NULL, *hex_data=NULL, *hex_path=NULL, *hex_host=NULL;
+	char *host=NULL, *port_str=NULL, *hostport;
+	char *http_auth=NULL;
 	char buf[8192], *retbuf=NULL;
 	int fd;
 	int sink=0;
+	int datalen_val;
+	int http_authlen=0;
 	unsigned int datalen_loc;
 
 	PRINTERR("Entering tcpcgi_getdata()");
@@ -134,12 +169,26 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 		return(NULL);
 	}
 
+	datalen_val=(datalen)?(*datalen):(0);
 	/* Need to free the hex_* variables past this point. */
 	hex_cmd=hexcode(cmd, strlen(cmd)); /* Guarenteed not NULL */
 	hex_session=hexcode(session, strlen(session)); /* Guarenteed not NULL */
 	hex_path=hexcode(http_path, strlen(http_path)); /* Guarenteed not NULL */
-	hex_data=hexcode(data, (datalen)?(*datalen):0);
-	hex_dest=hexcode(dest, strlen(dest?dest:""));
+	hex_data=hexcode(data, datalen_val);
+	if (dest) {
+		hostport=strdup(dest);
+		/* Parse the host:port value into two seperate variables */
+		port_str=strchr(hostport, ':');
+		if (!port_str) {
+			PRINTERR("Unable to parse destination.");
+			if (status) *status=-1;
+			return(NULL);
+		}
+		*port_str=0;
+		port_str++;
+		host=hostport;
+		hex_host=hexcode(host, strlen(host));
+	}
 
 	fd=createconnection_tcp(http_host, http_port);
 	if (fd<0) {
@@ -154,13 +203,24 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 		if (status) *status=-1;
 		return(NULL);
 	}
-	snprintf(buf, sizeof(buf), "cmd=%s&session=%s&dest=%s&data=%s&ts=%i&proto=%s", hex_cmd, hex_session, hex_dest, hex_data, (unsigned int) time(NULL), (proto_type==SOCK_DGRAM)?"udp":"tcp");
+
+	if (http_userpass) {
+		http_authlen=128+strlen(http_userpass);
+		http_auth=malloc(http_authlen);
+		snprintf(http_auth, http_authlen, "Authorization: Basic %s\r\n", http_userpass);
+	} else {
+		http_auth="";
+	}
+
+	snprintf(buf, sizeof(buf), "action=%s&id=%s&host=%s&port=%s&data=%s&datalen=%i&ts=%i&proto=%s", hex_cmd, hex_session, hex_host, port_str, hex_data, datalen_val, (unsigned int) time(NULL), (proto_type==SOCK_DGRAM)?"udp":"tcp");
+
+
 	buf[sizeof(buf)-1]='\0';
 	if (use_post) {
-		fprintf(fp, "POST %s HTTP/1.0\r\nHost: %s\r\nContent-length: %i\r\n\r\n%s\r\n", hex_path, http_host, (int) strlen(buf), buf);
+		fprintf(fp, "POST %s HTTP/1.0\r\nHost: %s\r\n%sContent-length: %i\r\n\r\n%s\r\n", hex_path, http_host, http_auth, (int) strlen(buf), buf);
 		PRINTERR("POST http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, (int) strlen(buf), buf);
 	} else {
-		fprintf(fp, "GET http://%s:%i%s?%s HTTP/1.0\r\n\r\n", http_host, http_port, hex_path, buf);
+		fprintf(fp, "GET %s?%s HTTP/1.0\r\nHost: %s\r\n%s\r\n", hex_path, buf, http_host, http_auth);
 		PRINTERR("GET http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, (int) strlen(buf), buf);
 	}
 	fflush(fp);
@@ -169,29 +229,40 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 	if (datalen) *datalen=0;
 	PRINTERR("Scheduling panic alarm.");
 	alarm(120); /* In case of emergency, please panic. */
-	if (fgets(buf, sizeof(buf), fp)==NULL) return(NULL);
+	if (fgets(buf, sizeof(buf), fp)==NULL) {
+		if (http_auth[0]!='\0') free(http_auth);
+		if (hex_cmd) free(hex_cmd);
+		if (hex_session) free(hex_session);
+		if (hex_data) free(hex_data);
+		if (hex_host) free(hex_host);
+		if (hex_path) free(hex_path);
+		fclose(fp);
+		return(NULL);
+	}
 	PRINTERR("Unscheduling panic alarm.");
 	alarm(0); /* Panic no more. */
 	if (strstr(buf," 40")) {
 		PRINTERR("File not available error, %s", buf);
+		if (http_auth[0]!='\0') free(http_auth);
 		if (hex_cmd) free(hex_cmd);
 		if (hex_session) free(hex_session);
 		if (hex_data) free(hex_data);
-		if (hex_dest) free(hex_dest);
+		if (hex_host) free(hex_host);
 		if (hex_path) free(hex_path);
 		fclose(fp);
 		return(NULL);
 	}
 	if (!strstr(buf, " 200 ")) {
 		PRINTERR("Invalid response, %s", buf);
+		if (http_auth[0]!='\0') free(http_auth);
 		if (hex_cmd) free(hex_cmd);
 		if (hex_session) free(hex_session);
 		if (hex_data) free(hex_data);
-		if (hex_dest) free(hex_dest);
+		if (hex_host) free(hex_host);
 		if (hex_path) free(hex_path);
 		sleep(1); /* Primitive backoff */
 		fclose(fp);
-		return(tcpcgi_getdata(http_host, http_port, http_path, cmd, session, dest, data, datalen, status, SOCK_STREAM));
+		return(tcpcgi_getdata(http_host, http_port, http_path, http_userpass, cmd, session, dest, data, datalen, status, SOCK_STREAM));
 	}
 	PRINTERR("Scheduling panic alarm.");
 	alarm(240); /* In case of emergency, please panic. */
@@ -203,7 +274,7 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 			sink=1; /* Only come here once. */
 			if (fgets(buf, sizeof(buf), fp)==NULL) break;
 			PRINTERR("statusmsg=%s", buf);
-			if (strstr(buf, "stat: ERROR:")!=NULL) {
+			if (strstr(buf, "stat: FAIL -")!=NULL) {
 				PRINTERR("Setting status to -ERR");
 				if (status) *status=-1;
 			}
@@ -226,10 +297,11 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 	alarm(0);
 
 	fclose(fp);
+	if (http_auth[0]!='\0') free(http_auth);
 	if (hex_cmd) free(hex_cmd);
 	if (hex_session) free(hex_session);
 	if (hex_data) free(hex_data);
-	if (hex_dest) free(hex_dest);
+	if (hex_host) free(hex_host);
 	if (hex_path) free(hex_path);
 	return(retbuf);
 }
@@ -254,7 +326,7 @@ void tcpcgi_closesession(int fd, char *http_host, int http_port, char *http_path
 	if (!session || fd<0) return;
 	PRINTERR("Terminating session \"%s\".", session);
 
-	tcpcgi_getdata(http_host, http_port, http_path, "close", session, NULL, NULL, NULL, NULL, 0);
+	tcpcgi_getdata(http_host, http_port, http_path, userpass, "close", session, NULL, NULL, NULL, NULL, 0);
 	return;
 }
 
@@ -281,14 +353,14 @@ int tcpcgi_handledata(char *http_host, int http_port, char *http_path, char *ses
 	}
 
 	if (lbuf && datalen!=0) {
-		cmd="put";
+		cmd="write";
 		dataretlen=datalen;
 	} else {
-		cmd="get";
+		cmd="read";
 		dataretlen=0;
 	}
 
-	ret=tcpcgi_getdata(http_host, http_port, http_path, cmd, session, dest, lbuf, &dataretlen, &status, proto_type);
+	ret=tcpcgi_getdata(http_host, http_port, http_path, userpass, cmd, session, dest, lbuf, &dataretlen, &status, proto_type);
 	PRINTERR("[%i](omitted, again)", dataretlen);
 	if (dataretlen>0 && ret && status>=0) {
 		write(fd, ret, dataretlen);
@@ -336,7 +408,7 @@ void sighandler(int sig) {
 
 int print_help(char *msg) {
 	fprintf(stderr, "Usage: tcpcgi -H host -f file -d dest [-P port] [-p port] [-m time]\n");
-	fprintf(stderr, "              [-x time] [-t time] [-b amt] [-rigluvh]\n");
+	fprintf(stderr, "              [-x time] [-t time] [-b amt] [-U user] [-rigluvh]\n");
 	fprintf(stderr, "    Options:\n");
 	fprintf(stderr, "         -H host      HTTP Server which has tcpcgi.cgi.\n");
 	fprintf(stderr, "         -f file      Full URL-relative on HTTP server to tcpcgi.cgi,\n");
@@ -356,6 +428,8 @@ int print_help(char *msg) {
 	fprintf(stderr, "                      should multiply the backoff time by this number,\n");
 	fprintf(stderr, "                      otherwise it's incremented by this amount (2 is the\n"); 
 	fprintf(stderr, "                      default).\n");
+	fprintf(stderr, "         -U user      Username and password to use for http authentication\n");
+	fprintf(stderr, "                      (must be in the form of user:pass).\n");
 	fprintf(stderr, "         -r           Specify that the backoff algorithm reset when remote\n");
 	fprintf(stderr, "                      activity occurs, instead of decrementing.\n");
 	fprintf(stderr, "         -i           Specify that the backoff algorithm should respond to\n");
@@ -414,7 +488,7 @@ int main(int argc, char **argv) {
 	pid_t chpid;
 #endif
 
-	while ((ch=getopt(argc, argv, "H:f:P:p:d:m:x:t:b:vgriulh"))!=-1) {
+	while ((ch=getopt(argc, argv, "H:f:P:p:d:m:x:t:b:U:vgriulh"))!=-1) {
 		switch (ch) {
 			case 'H':
 				http_host=strdup(optarg);
@@ -472,6 +546,9 @@ int main(int argc, char **argv) {
 			case 'u':
 				proto_type=(proto_type==SOCK_STREAM)?SOCK_DGRAM:SOCK_STREAM;
 				break;
+			case 'U':
+				userpass=mimecode(optarg, strlen(optarg));
+				break;
 			case ':':
 			case '?':
 			case 'h':
@@ -639,7 +716,7 @@ int main(int argc, char **argv) {
 				socks_info[pollpos].http_port=http_port;
 				socks_info[pollpos].http_host=http_host;
 				socks_info[pollpos].http_path=http_path;
-				tcpcgi_getdata(http_host, http_port, http_path, "open", socks_info[pollpos].session, hostport, NULL, NULL, &status, proto_type);
+				tcpcgi_getdata(http_host, http_port, http_path, userpass, "open", socks_info[pollpos].session, hostport, NULL, NULL, &status, proto_type);
 				if (status<0) {
 					PRINTERR("Error creating session on remote end.");
 					CLOSE_POLL(pollpos);
