@@ -29,7 +29,7 @@
 int main(void) {
 	int masterfd, clientfd;
 
-	while ((masterfd=createlisten(5011, 1))<0) { sleep(1); }
+	while ((masterfd=createlisten(5011, 1, SOCK_STREAM))<0) { sleep(1); }
 	while (1) {
 		clientfd=accept(masterfd, NULL, NULL);
 		dup2(clientfd, STDOUT_FILENO);
